"""
lineage.py – Compute all lineage instances and populate them
"""

import logging
from xuml_populate.config import mmdb
from typing import List, Set, Optional
from xuml_populate.tree.tree import extract
from xuml_populate.populate.mmclass_nt import Element_i, Spanning_Element_i, Lineage_i, Class_In_Lineage_i
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from pyral.relation import Relation

_logger = logging.getLogger(__name__)

tr_Lin = "Lineage"

class Lineage:
    """
    Create all lineages for a domain
    """

    domain = None

    lnums = 0
    walks = []
    xrels = set()
    xclasses = set()
    popclasses = set()
    lineages = None

    @classmethod
    def Derive(cls, domain: str):
        """

        :param domain:
        :return:
        """
        cls.domain = domain

        # Get all classes with at least one subclass facet and no superclass facets
        # These constitute 'leaves'. We use them as starting points as we step through a set of generalizations
        # to identify lineages.

        Relation.project(mmdb, attributes=('Class', 'Domain'), relation='Subclass', svar_name='subs')
        Relation.project(mmdb, attributes=('Class', 'Domain'), relation='Superclass', svar_name='supers')
        leaf_tuples = Relation.subtract(mmdb, rname1='subs', rname2='supers')
        leaf_classes = [t['Class'] for t in leaf_tuples.body]

        # Now we walk (step) through each generalization to build trees of one or more lineages
        for leaf in leaf_classes:
            cls.xrels = set()
            cls.xclasses = set()
            leafwalk = cls.step(walk=[], cvisit=leaf, rvisit=None)
            cls.walks.append(leafwalk)

        # We then prune these trees to extract unique branches, each of which constitutes a distinct lineage
        cls.lineages = set()
        for walk in cls.walks:
            pattern = walk.copy()
            if not any(isinstance(n, list) for n in pattern):
                pattern.sort()
                cls.lineages.add(':'.join(pattern))
            else:
                while len(pattern) > 0 and any(isinstance(n, list) for n in pattern):
                    extraction = extract(pattern)[0].copy()
                    extraction.sort()
                    cls.lineages.add(':'.join(extraction))

        # Finally, we load each lineage into the tclral
        cls.populate()

    @classmethod
    def step(cls, walk: List, cvisit: str, rvisit: Optional[str] = None) -> List:
        """
        Advance one step in the walk and return the result

        :param walk:  The walk prior to the step
        :param cvisit: The class currently being visited
        :param rvisit: The relationship currently being traversed
        :return: The updated walk
        """

        walk.append(cvisit)  # Advance the walk by adding the visited class
        cls.xclasses.add(cvisit)  # Class has been visited

        # Now we figure out where and if we can take another step

        # Get all adjacent relationships, if any, on the civisit class that have not already been traversed
        # Could be either superclass_name or subclasses, so we search Facets
        # Get all Facets that cvisit participates in
        R = f"Class:<{cvisit}>, Domain:<{cls.domain}>"
        Relation.restrict(mmdb, restriction=R, relation="Facet")
        s = Relation.project(mmdb, attributes=("Rnum", ))
        # Grab the result being careful to exclude prior traversals so we don't walk around in circles!
        adj_rels = [r['Rnum'] for r in s.body if r['Rnum'] not in cls.xrels and r['Rnum'] != rvisit]

        # We have nowhere else to walk if cvisit does not participate in any new rels
        if not adj_rels:
            return walk

        # Create a set of all hops going up to a superclass
        uphops = {h for h in adj_rels if cls.isSubclass(grel=h, cname=cvisit)}
        pass

        # We can try to take a step
        for arel in adj_rels:
            # Is cvisit a superclass or subclass in this arel?
            # There are only two possibilities, so we arbitrarily check to see if it paticipates as a subclass
            if arel in uphops:  # Hopping up to a superclass
                superclass = cls.findSuperclass(grel=arel)

                # Exclude all subclasses of unvisited uphop rels other than cvisit
                other_uphops = uphops - cls.xrels - {arel}
                for o in other_uphops:
                    subclasses = cls.findSubclasses(grel=o)
                    exclude_subclasses = {c for c in subclasses if c != cvisit}
                    cls.xclasses = cls.xclasses.union(exclude_subclasses)

                # Since we don't need to branch out, we can now mark this arel as excluded
                cls.xrels.add(arel)
                walk = cls.step(walk=walk, cvisit=superclass, rvisit=arel)
            else:  # hopping down to a subclass
                # We are going to branch out to one or more subclasses
                # (Any of our subclasses adjacent to some excluded relationship cannot be added)
                # Get all the subclass class names
                subclasses = cls.findSubclasses(grel=arel)
                visit_subs = subclasses.difference(cls.xclasses)
                for s in visit_subs:
                    # Start a new branch if there is more than one subclass to visit
                    fork = True if len(visit_subs) > 1 else False
                    if fork:
                        cls.xrels = set()  # New branch, no excluded rels
                        branch = cls.step(walk=[], cvisit=s, rvisit=arel)
                        if branch:
                            walk.append(branch)
                        else:
                            cls.xclasses.remove(s)
                    else:
                        walk = cls.step(walk=walk, cvisit=s, rvisit=arel)
                cls.xrels.add(arel)
        return walk

    @classmethod
    def findSubclasses(cls, grel: str) -> Set[str]:
        """
        Return the set of all subclasses in the specified generalization

        :param grel:
        :param domain:
        :return:
        """
        Relation.restrict(mmdb, relation='Subclass', restriction=f"Rnum:<{grel}>, Domain:<{cls.domain}>")
        s = Relation.project(mmdb, attributes=('Class', ))

        return {t['Class'] for t in s}

    @classmethod
    def isSubclass(cls, grel: str, cname: str) -> bool:
        Relation.restrict(mmdb,
                          relation='Subclass',
                          restriction=f"Class:<{cname}>, Rnum:<{grel}>, Domain:<{cls.domain}>")
        s = Relation.project(mmdb, attributes=())
        return bool(s.body)

    @classmethod
    def findSuperclass(cls, grel: str) -> str:
        """
        Traverse the specified relationship and return the name of the superclass

        :param grel:  A generalization relationship rnum
        :param domain:  A the name of the domain
        :return:
        """
        Relation.restrict(mmdb, relation='Superclass', restriction=f"Rnum:<{grel}>, Domain:<{cls.domain}>")
        s = Relation.project(mmdb, attributes=("Class", ))
        return s.body[0]['Class']

    @classmethod
    def populate(cls):
        """
        Trace through walks to populate all Lineages

        :return:
        """
        for lin in cls.lineages:
            cls.lnums += 1
            lnum = 'L' + (str(cls.lnums))
            _logger.info(f"Populating lineage [{lnum}]")
            Transaction.open(mmdb, tr_Lin)
            Relvar.insert(mmdb, tr=tr_Lin, relvar='Element', tuples=[
                Element_i(Label=lnum, Domain=cls.domain)
            ])
            Relvar.insert(mmdb, tr=tr_Lin, relvar='Spanning_Element', tuples=[
                Spanning_Element_i(Label=lnum, Domain=cls.domain)
            ])
            Relvar.insert(mmdb, tr=tr_Lin, relvar='Lineage', tuples=[
                Lineage_i(Lnum=lnum, Domain=cls.domain)
            ])
            for cname in lin.split(':'):
                Relvar.insert(mmdb, tr=tr_Lin, relvar='Class_In_Lineage', tuples=[
                    Class_In_Lineage_i(Class=cname,Lnum=lnum, Domain=cls.domain)
                ])
            Transaction.execute(mmdb, tr_Lin)



"""
rename_action.py – Populate a selection action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.ns_flow import NonScalarFlow
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mmclass_nt import Relational_Action_i, Table_Action_i, Rename_Action_i
from pyral.relvar import Relvar
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Rename = "Rename Action"

class RenameAction:
    """
    Create all relations for a Select Statement
    """

    @classmethod
    def populate(cls, input_nsflow: Flow_ap, from_attr: str, to_attr: str,
                 activity_data: Activity_ap ) -> (str, Flow_ap):
        """

        :param input_nsflow:
        :param from_attr:
        :param to_attr:
        :param activity_data:
        :return:
        """
        anum = activity_data.anum
        domain = activity_data.domain

        # Get header for the input flow
        table_header = NonScalarFlow.header(ns_flow=input_nsflow, domain=domain)
        # Rename the from attr keeping the same scalar
        from_scalar = table_header[from_attr]
        del table_header[from_attr]
        table_header[to_attr] = from_scalar
        # Create output table flow
        output_tflow = Flow.populate_relation_flow_by_header(table_header=table_header, anum=anum, domain=domain,
                                                             max_mult=input_nsflow.max_mult)

        # Create the action
        Transaction.open(db=mmdb, name=tr_Rename)
        action_id = Action.populate(tr=tr_Rename, anum=anum, domain=domain, action_type="rename")
        Relvar.insert(db=mmdb, tr=tr_Rename, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Rename, relvar='Table_Action', tuples=[
            Table_Action_i(ID=action_id, Activity=anum, Domain=domain, Input_a_flow=input_nsflow.fid,
                           Output_flow=output_tflow.fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Rename, relvar='Rename_Action', tuples=[
            Rename_Action_i(ID=action_id, Activity=anum, Domain=domain, From_attribute=from_attr,
                            From_non_scalar_type=input_nsflow.tname, To_attribute=to_attr, To_table=output_tflow.tname)
        ])
        Transaction.execute(db=mmdb, name=tr_Rename)
        return action_id, output_tflow



"""
project_action.py – Populate a Project Action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.exceptions.action_exceptions import ProjectedAttributeNotDefined
from scrall.parse.visitor import Projection_a
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.aparse_types import Flow_ap, Content, Activity_ap
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mmclass_nt import Relational_Action_i, Table_Action_i, Project_Action_i, \
    Projected_Attribute_i
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Project = "Project Action"


class ProjectAction:
    """
    Create all relations for a ProjectAction
    """

    @classmethod
    def populate(cls, input_nsflow: Flow_ap, projection: Projection_a, activity_data: Activity_ap) -> (str, Flow_ap):
        """
        Populate the Project Action

        :param input_nsflow: The input Non Scalar Flow that is being projected
        :param projection:  A list of attr names to be projected and optional expansion (ALL, EMPTY)
        :param activity_data:
        :return: Action ID and Projected table flow
        """
        # Save attribute values that we will need when creating the various select subsystem
        # classes
        domain = activity_data.domain
        anum = activity_data.anum

        _logger.info("Populating projection action")
        table_header = {}
        ns_type = input_nsflow.tname
        # Get type of each attribute
        for pattr in projection.attrs:
            if input_nsflow.content == Content.INSTANCE:
                R = f"Name:<{pattr.name}>, Class:<{input_nsflow.tname}>, Domain:<{domain}>"
                result = Relation.restrict(db=mmdb, relation='Attribute', restriction=R)
            else:
                R = f"Name:<{pattr.name}>, Table:<{input_nsflow.tname}>, Domain:<{domain}>"
                result = Relation.restrict(db=mmdb, relation='Table_Attribute', restriction=R)
            if not result.body:
                _logger.error(f"Attribute [{pattr.name}] in projection not defined on class [{input_nsflow.tname}]")
                raise ProjectedAttributeNotDefined
            table_header[pattr.name] = result.body[0]['Scalar']

        output_rel_flow = Flow.populate_relation_flow_by_header(table_header=table_header, anum=anum, domain=domain,
                                                                max_mult=input_nsflow.max_mult)

        Transaction.open(db=mmdb, name=tr_Project)
        action_id = Action.populate(tr=tr_Project, anum=anum, domain=domain, action_type="project")
        Relvar.insert(db=mmdb, tr=tr_Project, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Project, relvar='Table_Action', tuples=[
            Table_Action_i(ID=action_id, Activity=anum, Domain=domain, Input_a_flow=input_nsflow.fid,
                           Output_flow=output_rel_flow.fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Project, relvar='Project_Action', tuples=[
            Project_Action_i(ID=action_id, Activity=anum, Domain=domain)
        ])
        for pattr in projection.attrs:
            Relvar.insert(db=mmdb, tr=tr_Project, relvar='Projected_Attribute', tuples=[
                Projected_Attribute_i(Attribute=pattr.name, Non_scalar_type=ns_type, Project_action=action_id,
                                      Activity=anum, Domain=domain)
            ])
        Transaction.execute(db=mmdb, name=tr_Project)
        return action_id, output_rel_flow



"""
select_action.py – Populate a selection action instance in PyRAL
"""

import logging
from typing import List
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Activity_ap, Attribute_Comparison
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.expressions.restriction_condition import RestrictCondition
from xuml_populate.populate.mmclass_nt import (Select_Action_i, Single_Select_i, Identifier_Select_i,
                                               Zero_One_Cardinality_Select_i, Many_Select_i,
                                               Class_Restriction_Condition_i)
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Select = "Select Action"


class SelectAction:
    """
    Create all relations for a Select Statement
    """

    def __init__(self, input_instance_flow: Flow_ap, selection_parse, activity_data: Activity_ap):
        """

        :param input_instance_flow: The source flow into this selection
        :param selection_parse:  The parsed Scrall select action group
        :param activity_data:
        """
        self.input_instance_flow = input_instance_flow  # We are selecting instances from this instance flow
        self.selection_parse = selection_parse
        self.activity_data = activity_data
        self.domain = activity_data.domain
        self.anum = activity_data.anum

        self.attr_comparisons: List[Attribute_Comparison]

        self.expression = None
        self.comparison_criteria = []
        self.equivalence_criteria = []
        self.restriction_text = ""
        self.criterion_ctr = 0
        self.max_mult: MaxMult

        # Save attribute values that we will need when creating the various select subsystem
        # classes

        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Select)
        self.action_id = Action.populate(tr=tr_Select, anum=self.anum, domain=self.domain,
                                         action_type="select")  # Transaction open
        Relvar.insert(db=mmdb, tr=tr_Select, relvar='Select_Action', tuples=[
            Select_Action_i(ID=self.action_id, Activity=self.anum, Domain=self.domain,
                            Input_flow=self.input_instance_flow.fid)
        ])
        # Walk through the criteria parse tree storing any attributes or input flows
        # Also check to see if we are selecting on an identifier
        rcond = RestrictCondition(tr=tr_Select, action_id=self.action_id,
                                  input_nsflow=self.input_instance_flow,
                                  selection_parse=self.selection_parse,
                                  activity_data=self.activity_data
                                  )
        self.attr_comparisons = rcond.comparison_criteria
        self.selection_cardinality = rcond.cardinality
        self.sflows = rcond.input_scalar_flows

        Relvar.insert(db=mmdb, tr=tr_Select, relvar='Class_Restriction_Condition', tuples=[
            Class_Restriction_Condition_i(Select_action=self.action_id, Activity=self.anum, Domain=self.domain)
        ])
        # Create output flows
        self.max_mult, self.output_instance_flow = self.populate_multiplicity_subclasses()

        # We now have a transaction with all select-action instances, enter into the metamodel db
        Transaction.execute(db=mmdb, name=tr_Select)  # Select Action

    def identifier_selection(self) -> int:
        """
        | Determine whether we are selecting based on an identifier match.
        An identifier match supplies one value per identifier attribute for some identifier defined on
        the class.
         | Each comparison must be == (equivalence)
         |
         | **For example:**
         |
         | Assume Floor, Shaft is an identifier defined on the Accessible Shaft Level class
         | This means that if you supply a value for each attribute like so
         |
         | Accessible Shaft Level( Floor == x, Shaft == y )
         |
         | you will select at most one instance. But if you select based on:
         |
         | Accessible Shaft Level( Shaft == y )
         |
         | you may select multiple instances since a Shaft intersects multiple Floors
         :returns: An identifier number 1,2, ... or 0 if none found
        """
        idcheck = {c.attr for c in self.attr_comparisons if c.op == '=='}
        R = f"Class:<{self.input_instance_flow.tname}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Identifier_Attribute', restriction=R)
        Relation.project(db=mmdb, attributes=('Identifier', 'Attribute',), svar_name='all_id_attrs')
        # We have created a named relation with a projection of each id_attr and its id_num
        # Now we must step through each id_num to see if we are selecting on any of them
        i = 1  # Start with inum 1 {I}, (identifier 1). Every class has at least this identifier
        while True:
            # Step through every identifier of the class and see if there is a set of equivalence
            # comparisons that forms a superset of this identifier. If we are selecting at most one instance
            R = f"Identifier:<{str(i)}>"
            t_id_n_attrs = Relation.restrict(db=mmdb, relation='all_id_attrs', restriction=R)
            if not t_id_n_attrs.body:
                # This i num is not defined on the class, no more i nums to check
                break
            t_id_n_attr_names = Relation.project(db=mmdb, attributes=('Attribute',))
            id_n_attr_names = {t['Attribute'] for t in t_id_n_attr_names.body}
            if not id_n_attr_names - idcheck:
                # The set of identifier attributes for the current id number
                # is present in the set of attribute equivalence matches
                # So we are selecting on an identifier and at most one instance can flow out of the selection
                return i
            i += 1  # Increment to the next I num (I1, I2, etc)
        return 0

    def populate_multiplicity_subclasses(self) -> (MaxMult, Flow_ap):
        """
        Determine multiplicity of output and populate the relevant Select Action subclasses
        """
        # Determine if this should be an Identifier Select subclass that yields at most one instance
        selection_idnum = self.identifier_selection()
        if selection_idnum or self.selection_cardinality == 'ONE':
            max_mult = MaxMult.ONE
            # Populate a single instance flow for the selection output
            output_instance_flow = Flow.populate_instance_flow(
                cname=self.input_instance_flow.tname, anum=self.anum, domain=self.domain,
                label=None, single=True
            )
            _logger.info(f"INSERT Select action output single instance Flow: [{self.domain}:"
                         f"{self.input_instance_flow.tname}:{self.activity_data.activity_path.split(':')[-1]}"
                         f":{output_instance_flow}]")
            # Populate the Single Select subclass
            Relvar.insert(db=mmdb, tr=tr_Select, relvar='Single_Select', tuples=[
                Single_Select_i(ID=self.action_id, Activity=self.anum, Domain=self.domain,
                                Output_flow=output_instance_flow.fid)
            ])
            if selection_idnum:
                # Populate an Identifier Select subclass
                Relvar.insert(db=mmdb, tr=tr_Select, relvar='Identifier_Select', tuples=[
                    Identifier_Select_i(ID=self.action_id, Activity=self.anum, Domain=self.domain,
                                        Identifier=selection_idnum, Class=self.input_instance_flow.tname)
                ])
            else:
                # Populate an Identifier Select subclass
                # Note that if both ONE cardinality specified and identifier select, identifier select takes precedence
                Relvar.insert(db=mmdb, tr=tr_Select, relvar='Zero_One_Cardinality_Select', tuples=[
                    Zero_One_Cardinality_Select_i(ID=self.action_id, Activity=self.anum, Domain=self.domain)
                ])
        else:
            # Many select with Multiple Instance Flow output
            max_mult = MaxMult.MANY
            output_instance_flow = Flow.populate_instance_flow(
                cname=self.input_instance_flow.tname, anum=self.anum,
                domain=self.domain, label=None, single=False
            )
            _logger.info(f"INSERT Select action output multiple instance Flow: [{self.domain}:"
                         f"{self.input_instance_flow.tname}:{self.activity_data.activity_path.split(':')[-1]}"
                         f":{output_instance_flow}]")
            # Populate the Many Select subclass
            Relvar.insert(db=mmdb, tr=tr_Select, relvar='Many_Select', tuples=[
                Many_Select_i(ID=self.action_id, Activity=self.anum, Domain=self.domain,
                              Output_flow=output_instance_flow.fid)
            ])
        return max_mult, output_instance_flow



"""
reference.py – Functionality to manage referential attributes
"""

# System
import logging

# Model Integration
from pyral.relation import Relation

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.exceptions.class_exceptions import MixedTargetID, ReferenceToNonIdentifier

_logger = logging.getLogger(__name__)

def targetid(ref, rnum: str, domain: str) -> int:
    # OR restriction criteria not yet supported in PyRAL, so we iterate on the attributes
    to_id = None
    for to_attr in ref['attrs']:
        R = f"Attribute:<{to_attr}>, Class:<{ref['class']}>, Domain:<{domain}>"
        result = Relation.restrict(db=mmdb, relation='Identifier_Attribute', restriction=R)
        if not result.body:
            _logger.exception(f"No identifier found in attribute reference on [{rnum}]")
            raise ReferenceToNonIdentifier
        attr_id = int(result.body[0]['Identifier'])
        if not to_id:
            to_id = attr_id
        elif to_id != attr_id:
            _logger.exception(f"Multiple identifiers referenced on [{rnum}]")
            raise MixedTargetID
    return to_id

